"""
Test calibration
"""

#%% Imports and settings
import sciris as sc
import starsim as ss
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from ax.plot.contour import plot_contour
from ax.plot.trace import optimization_trace_single_method
from ax.service.managed_loop import optimize
from ax.utils.notebook.plotting import init_notebook_plotting, render

do_plot = 1
do_save = 0
n_agents = 2e3

#%% Helper functions

def make_sim():
    sir = ss.SIR(
        beta = ss.beta(0.9),
        dur_inf = ss.lognorm_ex(mean=ss.dur(6)),
        init_prev = ss.bernoulli(0.01),
    )

    #deaths = ss.Deaths(death_rate=15)
    #births = ss.Births(birth_rate=15)

    random = ss.RandomNet(n_contacts=ss.poisson(4))

    sim = ss.Sim(
        dt = 1,
        unit = 'day',
        n_agents = n_agents,
        #total_pop = 9980999,
        start = sc.date('2024-01-01'),
        stop = sc.date('2024-01-31'),
        diseases = sir,
        networks = random,
        #demographics = [deaths, births],
    )

    return sim


def build_sim(sim, calib_pars, **kwargs):
    """ Modify the base simulation by applying calib_pars """

    for k, v in calib_pars.items():
        if k == 'beta':
            sim.diseases.sir.pars['beta'] = ss.beta(v)
        elif k == 'dur_inf':
            sim.diseases.sir.pars['dur_inf'] = ss.lognorm_ex(mean=ss.dur(v)), #ss.dur(v)
        elif k == 'n_contacts':
            sim.networks.randomnet.pars.n_contacts = v # Typically a Poisson distribution, but this should set the distribution parameter value appropriately
        else:
            sim.pars[k] = v # Assume sim pars

    return sim

def eval_sim(pars):
    sim = make_sim()
    sim.init()
    sim = build_sim(sim, pars)
    sim.run()
    #print('pars:', pars, ' --> Final prevalence:', sim.results.sir.prevalence[-1])
    fig = sim.plot()
    fig.suptitle(pars)
    fig.subplots_adjust(top=0.9)
    plt.show()

    return dict(
        prevalence_error = ((sim.results.sir.prevalence[-1] - 0.10)**2, None),
        prevalence = (sim.results.sir.prevalence[-1], None),
    )


#%% Define the tests
def test_calibration(do_plot=False):
    sc.heading('Testing calibration')

    # Define the calibration parameters
    calib_pars = [
        dict(name='beta', type='range', bounds=[0.01, 1.0], value_type='float', log_scale=True),
        dict(name='dur_inf', type='range', bounds=[1, 60], value_type='float', log_scale=False),
        #dict(name='init_prev', type='range', bounds=[0.01, 0.30], value_type='float', log_scale=False),
        dict(name='n_contacts', type='range', bounds=[2, 10], value_type='int', log_scale=False),
    ]

    best_pars, values, exp, model = optimize(
        experiment_name = 'starsim',
        parameters = calib_pars,
        evaluation_function = eval_sim,
        objective_name = 'prevalence_error',
        minimize = True,
        parameter_constraints = None,
        outcome_constraints = None,
        total_trials = 10,
        arms_per_trial = 3,
    )

    return best_pars, values, exp, model


#%% Run as a script
if __name__ == '__main__':

    T = sc.timer()
    do_plot = True

    best_pars, values, exp, model = test_calibration(do_plot=do_plot)

    print('best_pars:', best_pars)
    print('values:', values)
    print('exp:', exp)
    print('model:', model)

    render(plot_contour(model=model, param_x='beta', param_y='init_prev', metric_name='prevalence'))

    # `plot_single_method` expects a 2-d array of means, because it expects to average means from multiple
    # optimization runs, so we wrap out best objectives array in another array.

    for trial in exp.trials.values():
        print(trial)
        print(dir(trial))
        print(f"Trial {trial.index} with parameters {trial.arm.parameters} "
              f"has objective {trial.objective_mean}.") 

    best_objectives = np.array(
        [[trial.objective_mean for trial in exp.trials.values()]]
    )
    best_objective_plot = optimization_trace_single_method(
        y = np.minimum.accumulate(best_objectives, axis=1),
        optimum = 0.10, #hartmann6.fmin,
        title = "Model performance vs. # of iterations",
        ylabel = "Prevalence",
    )
    render(best_objective_plot)

    plt.show()

    T.toc()



"""
Functions and classes for handling time
"""
import copy
import sciris as sc
import numpy as np
import pandas as pd
import datetime as dt
import dateutil as du
import starsim as ss


#%% Helper functions

# Classes and objects that are externally visible
__all__ = ['time_units', 'time_ratio', 'date_add', 'date_diff']

# Allowable time arguments
time_args = ['start', 'stop', 'dt', 'unit']

# Define available time units
time_units = sc.objdict(
    day   = 1.0,
    week  = 7.0,
    month = 30.4375, # 365.25/12
    year  = 365.25, # For consistency with months
)

# Define defaults
default_dur = 50
default_unit = 'year'
default_start_year = 2000
default_start_date = '2000-01-01'
default_start = sc.objdict(
    {k:default_start_date for k in ['day', 'week', 'month']} |
    {k:default_start_year for k in ['year', 'unitless']}
)


def time_ratio(unit1='day', dt1=1.0, unit2='day', dt2=1.0, as_int=False):
    """
    Calculate the relationship between two sets of time factors

    Args:
        unit1 (str): units for the numerator
        dt1 (float): timestep for the numerator
        unit2 (str): units for the denominator
        dt2 (float): timestep for the denominator
        as_int (bool): round and convert to an integer

    **Example**::
        ss.time_ratio(unit1='week', dt1=2, unit2='day', dt2=1, as_int=True) # Returns 14
    """
    if dt1 == dt2:
        dt_ratio = 1.0
    else:
        if dt1 is None or dt2 is None:
            errormsg = f'Cannot convert between dt when one is None ({dt1}, {dt2})'
            raise ValueError(errormsg)
        dt_ratio = dt1/dt2

    if unit1 == unit2:
        unit_ratio = 1.0
    else:
        if is_unitless(unit1) or is_unitless(unit2):
            errormsg = f'Cannot convert between units if only one is unitless (unit1={unit1}, unit2={unit2})'
            raise ValueError(errormsg)
        if unit1 is None or unit2 is None:
            errormsg = f'Cannot convert between units when only one has been initialized (unit1={unit1}, unit2={unit2})'
            raise ValueError(errormsg)
        u1 = time_units[unit1]
        u2 = time_units[unit2]
        unit_ratio = u1/u2

    factor = dt_ratio * unit_ratio
    if as_int:
        factor = int(round(factor))
    return factor


def date_add(start, dur, unit):
    """ Add two dates (or integers) together """
    if sc.isnumber(start):
        stop = start + dur
    else:
        if unit in time_units:
            ndays = int(round(time_units[unit]*dur))
            stop = sc.datedelta(start, days=ndays)
        else:
            errormsg = f'Unknown unit {unit}, choices are: {sc.strjoin(time_units.keys())}'
            raise ValueError(errormsg)
    return stop


def date_diff(start, stop, unit):
    """ Find the difference between two dates (or integers) """
    if sc.isnumber(start) and sc.isnumber(stop):
        dur = stop - start
    else:
        if unit == 'year':
            dur = sc.datetoyear(stop) - sc.datetoyear(start) # TODO: allow non-integer amounts
        else:
            dur = (ss.date(stop) - ss.date(start)).days
            if unit != 'day':
                dur *= time_ratio(unit1='day', unit2=unit)
    return dur


def round_tvec(tvec):
    """ Round time vectors to a certain level of precision, to avoid floating point errors """
    decimals = int(-np.log10(ss.options.time_eps))
    tvec = np.round(np.array(tvec), decimals=decimals)
    return tvec


def years_to_dates(yearvec):
    """ Convert a numeric year vector to a date vector """
    datevec = np.array([date(sc.datetoyear(y, reverse=True)) for y in yearvec])
    return datevec


def dates_to_years(datevec):
    """ Convert a date vector to a numeric year vector"""
    yearvec = round_tvec([sc.datetoyear(d.date()) for d in datevec])
    return yearvec


def dates_to_days(datevec, start_date):
    """ Convert a date vector into relative days since start date """
    start_date = date(start_date)
    dayvec = np.array([(d - start_date).days for d in datevec])
    return dayvec


#%% Time classes

__all__ += ['date', 'Time']


class date(pd.Timestamp):
    """
    Define a single date; based on ``pd.Timestamp``

    Args:
        date (int/float/str/datetime): Any type of date input (ints and floats will be interpreted as years)
        kwargs (dict): passed to pd.Timestamp()

    **Examples**::

        ss.date(2020) # Returns <2020-01-01>
        ss.date(year=2020) # Returns <2020-01-01>
        ss.date(year=2024.75) # Returns <2024-10-01>
        ss.date('2024-04-04') # Returns <2024-04-04>
        ss.date(year=2024, month=4, day=4) # Returns <2024-04-04>
    """
    def __new__(cls, *args, **kwargs):
        """ Check if a year was supplied, and preprocess it; complex due to pd.Timestamp implementation """
        single_year_arg = False
        if len(args) == 1:
            arg = args[0]
            if arg is None:
                return pd.Timestamp(None)
            elif sc.isnumber(arg):
                single_year_arg = True
            elif isinstance(arg, pd.Timedelta):
                return pd.Timedelta(arg)
        year_kwarg = len(args) == 0 and len(kwargs) == 1 and 'year' in kwargs
        if single_year_arg:
            return cls.from_year(args[0])
        if year_kwarg:
            return cls.from_year(kwargs['year'])

        # Otherwise, proceed as normal
        out = super(date, cls).__new__(cls, *args, **kwargs)
        out = cls._reset_class(out)
        return out

    def __copy__(self):
        """ Required due to pd.Timestamp implementation; pd.Timestamp is immutable, so create new object """
        out = self.__class__(self)
        return out

    def __deepcopy__(self, *args, **kwargs):
        """ Required due to pd.Timestamp implementation; pd.Timestamp is immutable, so create new object """
        out = self.__class__(self)
        return out

    @classmethod
    def _reset_class(cls, obj):
        """ Manually reset the class from pd.Timestamp to ss.date """
        try:
            obj.__class__ = date
        except:
            warnmsg = f'Unable to convert {obj} to ss.date(); proceeding with pd.Timestamp'
            ss.warn(warnmsg)
        return obj

    def __repr__(self, bracket=True):
        """ Show the date in brackets, e.g. <2024.04.04> """
        _ = ss.options.date_sep
        y = f'{self.year:04d}'
        m = f'{self.month:02d}'
        d = f'{self.day:02d}'
        string = y + _ + m + _ + d
        if bracket:
            string = '<' + string + '>'
        return string

    def __str__(self):
        """ Like repr, but just the date, e.g. 2024.04.04 """
        return self.__repr__(bracket=False)

    def replace(self, *args, **kwargs):
        """ Returns a new ss.date(); pd.Timestamp is immutable """
        out = super().replace(*args, **kwargs)
        out = self.__class__._reset_class(out)
        return out

    def disp(self, **kwargs):
        """ Show the full object """
        return sc.pr(self, **kwargs)

    @classmethod
    def from_year(cls, year):
        """
        Convert an int or float year to a date.

        **Examples**::

            ss.date.from_year(2020) # Returns <2020-01-01>
            ss.date.from_year(2024.75) # Returns <2024-10-01>
        """
        if isinstance(year, int):
            return cls(year=year, month=1, day=1)
        else:
            dateobj = sc.datetoyear(year, reverse=True)
            return cls(dateobj)

    def to_year(self):
        """
        Convert a date to a floating-point year

        **Examples**::

            ss.date('2020-01-01').to_year() # Returns 2020.0
            ss.date('2024-10-01').to_year() # Returns 2024.7486
        """
        return sc.datetoyear(self.date())

    def to_pandas(self):
        """ Convert to a standard pd.Timestamp instance """
        return pd.Timestamp(self.to_numpy()) # Need to convert to NumPy first or it doesn't do anything

    @staticmethod
    def _convert_other(other):
        if isinstance(other, du.relativedelta.relativedelta):
            tu = ss.time.time_units
            days = other.days + tu.month*other.months + tu.year*other.years
            int_days = int(round(days))
            other = dt.timedelta(days=int_days)
        elif isinstance(other, ss.dur):
            factor = ss.time.time_units[other.unit]
            int_days = int(round(factor*other.v))
            other = dt.timedelta(days=int_days)
        return other

    def __add__(self, other):
        other = self._convert_other(other)
        out = super().__add__(other)
        out = date(out)
        return out

    def __sub__(self, other):
        other = self._convert_other(other)
        out = super().__sub__(other)
        out = date(out)
        return out

    def __radd__(self, other): return self.__add__(other)
    def __iadd__(self, other): return self.__add__(other)
    def __rsub__(self, other): return self.__sub__(other) # TODO: check if this should be reversed
    def __isub__(self, other): return self.__sub__(other)


class Time(sc.prettyobj):
    """
    Handle time vectors for both simulations and modules.

    Args:
        start (float/str/date): the start date for the simulation/module
        stop (float/str/date): the end date for the simulation/module
        dt (float): the step size, in units of "unit"
        unit (str): the time unit; choices are "day", "week", "month", "year", or "unitless"
        pars (dict): if provided, populate parameter values from this dictionary
        parent (obj): if provided, populate missing parameter values from a 'parent" ``Time`` instance
        name (str): if provided, name the ``Time`` object
        init (bool): whether or not to immediately initialize the Time object
        sim (bool/Sim): if True, initializes as a sim-specific ``Time`` instance; if a Sim instance, initialize the absolute time vector

    The ``Time`` object, after initialization, has the following attributes:

    - ``ti`` (int): the current timestep
    -  ``dt_year`` (float): the timestep in units of years
    - ``npts`` (int): the number of timesteps
    - ``tvec`` (array): time starting at 0, in self units (e.g. ``[0, 0.1, 0.2, ... 10.0]`` if start=0, stop=10, dt=0.1)
    - ``absvec`` (array): time relative to sim start, in units of sim units (e.g. ``[366, 373, 380, ...]`` if sim-start=2001, start=2002, sim-unit='day', unit='week')
    - ``yearvec`` (array): time represented as floating-point years (e.g. ``[2000, 2000.1, 2000.2, ... 2010.0]`` if start=2000, stop=2010, dt=0.1)
    - ``datevec`` (array): time represented as an array of ``ss.date`` objects (e.g. ``[<2000.01.01>, <2000.02.07>, ... <2010.01.01>]`` if start=2000, stop=2010, dt=0.1)
    - ``timevec`` (array): the "native" time vector, which always matches one of ``tvec``, ``yearvec``, or ``datevec``

    **Examples**::

        t1 = ss.Time(start=2000, stop=2020, dt=1.0, unit='year') # Years, numeric units
        t2 = ss.Time(start='2021-01-01', stop='2021-04-04', dt=2.0, unit='day') # Days, date units
    """
    def __init__(self, start=None, stop=None, dt=None, unit=None, pars=None, parent=None,
                 name=None, init=True, sim=None):
        self.name = name
        self.start = start
        self.stop = stop
        self.dt = dt
        self.unit = unit
        self.ti = 0 # The time index, e.g. 0, 1, 2

        # Prepare for later initialization
        self.dt_year = None
        self.npts    = None
        self.tvec    = None
        self.timevec = None
        self.datevec = None
        self.yearvec = None
        self.abstvec = None
        self.initialized = False

        # Finalize
        self.update(pars=pars, parent=parent)
        if init and self.ready:
            self.init(sim=sim)
        return

    def __bool__(self):
        """ Always truthy """
        return True

    def __len__(self):
        """ Length is the number of timepoints """
        return sc.ifelse(self.npts, 0)

    @property
    def ready(self):
        """ Check if all parameters are in place to be initialized """
        return not any([getattr(self, k) is None for k in time_args])

    @property
    def is_numeric(self):
        """ Check whether the fundamental simulation unit is numeric (as opposed to date-based) """
        try:
            return sc.isnumber(self.start)
        except:
            return False

    @property
    def is_unitless(self):
        return is_unitless(self.unit)

    def __setstate__(self, state):
        """ Custom setstate to unpickle ss.date() instances correctly """
        self.__dict__.update(state)
        self._convert_timestamps()
        return

    def _convert_timestamps(self):
        """ Replace pd.Timestamp instances with ss.date(); required due to pandas limitations with pickling """
        objs = [self.start, self.stop]
        objs += sc.tolist(self.datevec, coerce='full')
        objs += sc.tolist(self.timevec, coerce='full')
        objs = [obj for obj in objs if type(obj) == pd.Timestamp]
        for obj in objs:
            date._reset_class(obj)
        return

    def update(self, pars=None, parent=None, reset=True, force=None, **kwargs):
        """ Reconcile different ways of supplying inputs """
        pars = sc.mergedicts(pars)
        stale = False

        for key in time_args:
            current_val = getattr(self, key, None)
            parent_val = getattr(parent, key, None)
            kw_val = kwargs.get(key)
            par_val = pars.get(key)

            # Special handling for dt: don't inherit dt if the units are different
            if key == 'dt':
                if isinstance(parent, Time):
                    if parent.unit != self.unit:
                        parent_val = 1.0

            if force is False: # Only update missing (None) values
                val = sc.ifelse(current_val, kw_val, par_val, parent_val)
            elif force is None: # Prioritize current value
                val = sc.ifelse(kw_val, par_val, current_val, parent_val)
            elif force is True: # Prioritize parent value
                val = sc.ifelse(kw_val, par_val, parent_val, current_val)
            else:
                errormsg = f'Invalid value {force} for force: must be False, None, or True'
                raise ValueError(errormsg)

            if val != current_val:
                setattr(self, key, val)
                stale = True

        if stale and reset and self.initialized:
            self.init()
        return

    def init(self, sim=None):
        """ Initialize all vectors """
        # Initial validation
        self.unit = validate_unit(self.unit)

        # Copy missing values from sim
        if isinstance(sim, ss.Sim):
            self.unit = sc.ifelse(self.unit, sim.t.unit)
            if self.unit == sim.t.unit: # Units match, use directly
                sim_dt = sim.t.dt
                sim_start = sim.t.start
                sim_stop = sim.t.stop
            else: # Units don't match, use datevec instead
                sim_dt = 1.0 # Don't try to reset the dt if the units don't match
                sim_start = sim.t.datevec[0]
                sim_stop = sim.t.datevec[-1]
            self.dt = sc.ifelse(self.dt, sim_dt)
            self.start = sc.ifelse(self.start, sim_start)
            self.stop = sc.ifelse(self.stop, sim_stop)

        # Handle start and stop
        self.start = self.start if self.is_numeric else date(self.start)
        self.stop  = self.stop  if self.is_numeric else date(self.stop)

        # Convert start and stop to dates
        date_start = self.start
        date_stop = self.stop
        date_unit = 'year' if not has_units(self.unit) else self.unit # Use year by default
        dt_year = time_ratio(unit1=date_unit, dt1=self.dt, unit2='year', dt2=1.0) # Timestep in units of years
        offset = 0
        if self.is_numeric and date_start == 0:
            date_start = ss.date(ss.time.default_start[date_unit])
            date_stop = date_start + ss.dur(date_stop, unit=date_unit)
            offset = date_start.year

        # If numeric, treat that as the ground truth
        if self.is_numeric:
            ratio = time_ratio(unit1=date_unit, unit2='year')
            timevec = round_tvec(sc.inclusiverange(self.start, self.stop, self.dt))
            yearvec = round_tvec((timevec-timevec[0])*ratio + offset + timevec[0]) # TODO: simplify
            datevec = years_to_dates(yearvec)

        # If unitless, just use that
        elif self.is_unitless:
            timevec = round_tvec(sc.inclusiverange(self.start, self.stop, self.dt))
            yearvec = round_tvec(timevec)
            datevec = timevec

        # If the unit is years, handle that
        elif date_unit == 'year': # For years, the yearvec is the most robust representation
            start_year = sc.datetoyear(date_start.date())
            stop_year = sc.datetoyear(date_stop.date())
            yearvec = round_tvec(sc.inclusiverange(start_year, stop_year, self.dt))
            datevec = years_to_dates(yearvec)
            timevec = datevec

        # Otherwise, use dates as the ground truth
        else:
            if int(self.dt) == self.dt: # The step is integer-like, use exactly
                key = date_unit + 's' # e.g. day -> days
                datelist = sc.daterange(date_start, date_stop, interval={key:int(self.dt)})
            else: # Convert to the sim unit instead
                day_delta = time_ratio(unit1=date_unit, dt1=self.dt, unit2='day', dt2=1.0, as_int=True)
                if day_delta >= 1:
                    datelist = sc.daterange(date_start, date_stop, interval={'days':day_delta})
                else:
                    errormsg = f'Timestep {dt} is too small; must be at least 1 day'
                    raise ValueError(errormsg)
            datevec = np.array([ss.date(d) for d in datelist])
            yearvec = dates_to_years(datevec)
            timevec = datevec

        # Store things
        self.dt_year = dt_year
        self.npts = len(timevec) # The number of points in the sim
        self.tvec = round_tvec(np.arange(self.npts)*self.dt) # Absolute time array
        self.timevec = timevec
        self.datevec = datevec
        self.yearvec = yearvec
        if sim == True: # It's the sim itself, the tvec is the absolute time vector
            self.abstvec = self.tvec
        elif sim is not None:
            self.make_abstvec(sim)
        else:
            self.abstvec = None # Intentionally set to None, cannot be used in the sim loop until populated
        self.initialized = True
        return

    def make_abstvec(self, sim):
        """ Convert the current time vector into sim units """
        # Validation
        if self.is_unitless != sim.t.is_unitless:
            errormsg = f'Cannot mix units with unitless time: sim.unit={sim.t.unit} {self.name}.unit={self.unit}'
            raise ValueError(errormsg)

        # Both are unitless or numeric
        both_unitless = self.is_unitless and sim.t.is_unitless
        both_numeric = self.is_numeric and sim.t.is_numeric
        if both_unitless or both_numeric:
            abstvec = self.tvec.copy() # Start by copying the current time vector
            ratio = time_ratio(unit1=self.unit, dt1=1.0, unit2=sim.t.unit, dt2=1.0) # tvec has sim units, but not dt
            if ratio != 1.0:
                abstvec *= ratio # TODO: CHECK
            start_diff = self.start - sim.t.start
            if start_diff != 0.0:
                abstvec += start_diff

        # The sim uses years; use yearvec
        elif sim.t.unit == 'year':
            abstvec = self.yearvec.copy()
            abstvec -= sim.t.yearvec[0] # Start relative to sim start

        # Otherwise (days, weeks, months), use datevec and convert to days
        else:
            dayvec = dates_to_days(self.datevec, start_date=sim.t.datevec[0])
            ratio = time_ratio(unit1='day', dt1=1.0, unit2=sim.t.unit, dt2=1.0)
            abstvec = dayvec*ratio # Convert into sim time units

        self.abstvec = round_tvec(abstvec) # Avoid floating point inconsistencies
        return

    def now(self, key=None):
        """
        Get the current simulation time

        Args:
            which (str): which type of time to get: default (None), "year", "date", "tvec", or "str"

        **Examples**::

            t = ss.Time(start='2021-01-01', stop='2022-02-02', dt=1, unit='week')
            t.ti = 25
            t.now() # Returns <2021-06-25>
            t.now('date') # Returns <2021-06-25>
            t.now('year') # Returns 2021.479
            t.now('str') # Returns '2021-06-25'
        """
        if key in [None, 'none', 'time', 'str']:
            vec = self.timevec
        elif key == 'tvec':
            vec = self.tvec
        elif key == 'date':
            vec = self.datevec
        elif key == 'year':
            vec = self.yearvec
        else:
            errormsg = f'Invalid key "{key}": must be None, abs, date, or year'
            raise ValueError(errormsg)

        ti = np.clip(self.ti, 0, len(vec)-1)
        now = vec[ti]
        if key == 'str':
            now = f'{now:0.1f}' if isinstance(now, float) else str(now)
        return now


#%% TimePar classes

__all__ += ['TimePar', 'dur', 'days', 'years', 'rate', 'perday', 'peryear',
            'time_prob', 'beta', 'rate_prob']

class TimePar(ss.BaseArr):
    """
    Base class for time-aware parameters, durations and rates

    NB, because the factor needs to be recalculated, do not set values directly.
    """
    def __new__(cls, v=None, *args, **kwargs):
        """ Allow TimePars to wrap distributions and return the distributions """

        # Special distribution handling
        if isinstance(v, ss.Dist):
            if len(args):
                errormsg = f'When wrapping a distribution with a TimePar, args not allowed ({args}); use kwargs'
                raise ValueError(errormsg)
            dist = v
            dist.pars[0] = cls(dist.pars[0], **kwargs) # Convert the first parameter to a TimePar (the same scale is applied to all parameters)
            return dist

        # Otherwise, do the usual initialization
        else:
            return super().__new__(cls)

    def __init__(self, v, unit=None, parent_unit=None, parent_dt=None, self_dt=1.0):
        self.v = v
        self.unit = unit
        self.parent_unit = parent_unit
        self.parent_dt = parent_dt
        self.self_dt = self_dt
        self.factor = None
        self.values = None
        self.initialized = False
        self.validate_units()
        return

    def validate_units(self):
        """ Check that the units entered are valid """
        try:
            self.unit = unit_mapping[self.unit]
        except KeyError:
            errormsg = f'Invalid unit "{self.unit}"; must be one of: {sc.strjoin(time_units.keys())}'
            raise ValueError(errormsg)
        try:
            self.parent_unit = unit_mapping[self.parent_unit]
        except KeyError:
            errormsg = f'Invalid parent unit "{self.parent_unit}"; must be one of: {sc.strjoin(time_units.keys())}'
            raise ValueError(errormsg)
        return

    def init(self, parent=None, parent_unit=None, parent_dt=None, update_values=True, die=True):
        """ Link to the sim and/or module units """
        if parent is None:
            parent = sc.dictobj(unit=parent_unit, dt=parent_dt)
        else:
            if parent_dt is not None:
                errormsg = f'Cannot override parent {parent} by setting parent_dt; set in parent object instead'
                raise ValueError(errormsg)

        if parent.unit is not None:
            self.parent_unit = parent.unit

        if parent.dt is not None:
            self.parent_dt = parent.dt

        # Set defaults if not yet set
        self.unit = sc.ifelse(self.unit, self.parent_unit) # If unit isn't defined but parent is, set to parent
        self.parent_unit = sc.ifelse(self.parent_unit, self.unit) # If parent isn't defined but unit is, set to self
        self.parent_dt = sc.ifelse(self.parent_dt, self.self_dt, 1.0) # If dt isn't defined, assume 1 (self_dt is defined by default)

        # Calculate the actual conversion factor to be used in the calculations
        self.update_cached(update_values=update_values, die=die)
        self.initialized = True
        self.validate_units()
        return self

    def __repr__(self):
        name = self.__class__.__name__

        if self.initialized:
            if self.factor == 1.0:
                xstr = ''
            else:
                xstr = f', values={self.values}'
        else:
            xstr = ', initialized=False'

        if (self.parent_unit is not None) and (self.unit != self.parent_unit):
            parentstr = f', parent={self.parent_unit}'
        else:
            parentstr = ''

        default_dt = sc.ifelse(self.self_dt, 1.0) == 1.0
        if not default_dt:
            dtstr = f', self_dt={self.self_dt}'
        else:
            dtstr = ''

        # Rather than ss.dur(3, unit='day'), dispaly as ss.days(3)
        prefixstr = 'ss.'
        key = (name, self.unit)
        mapping = {
            ('dur',  'day'):  'days',
            ('dur',  'year'): 'years',
            ('rate', 'day'):  'perday',
            ('rate', 'year'): 'peryear',
        }

        if key in mapping and default_dt:
            prefixstr += mapping[key]
            unitstr = ''
        else:
            prefixstr += name
            unitstr = f', unit={self.unit}'

        suffixstr = unitstr + parentstr + dtstr + xstr

        return f'{prefixstr}({self.v}{suffixstr})'

    @property
    def isarray(self):
        """ Check if the value is an array """
        return isinstance(self.v, np.ndarray)

    def set(self, v=None, unit=None, parent_unit=None, parent_dt=None, self_dt=None, force=False):
        """ Set the specified parameter values (ignoring None values) and update stored values """
        if v           is not None: self.v           = v
        if unit        is not None: self.unit        = unit
        if parent_unit is not None: self.parent_unit = parent_unit
        if parent_dt   is not None: self.parent_dt   = parent_dt
        if self_dt     is not None: self.self_dt     = self_dt
        if self.initialized or force: # Don't try to set these unless it's been initialized
            self.update_cached()
        self.validate_units()
        return self

    def update_cached(self, update_values=True, die=True):
        """ Update the cached factor and values """
        try:
            self.update_factor()
            if update_values:
                self.update_values()
        except TypeError as E: # For a known error, skip silently if die=False
            if die:
                errormsg = f'Update failed for {self}. Argument v={self.v} should be a number or array; if a function, use update_values=False. Error: {E}'
                raise TypeError(errormsg) from E
        except Exception as E: # For other errors, raise a warning
            if die:
                raise E
            else:
                tb = sc.traceback(E)
                warnmsg = f'Uncaught error encountered while updating {self}, but die=False. Traceback:\n{tb}'
                ss.warn(warnmsg)

        return self

    def update_factor(self):
        """ Set factor used to multiply the value to get the output """
        self.factor = time_ratio(unit1=self.unit, dt1=self.self_dt, unit2=self.parent_unit, dt2=self.parent_dt)
        return

    def update_values(self):
        """ Convert from self.v to self.values based on self.factor -- must be implemented by derived classes """
        raise NotImplementedError

    def to(self, unit=None, dt=None):
        """ Create a new timepar based on the current one but with a different unit and/or dt """
        new = self.asnew()
        unit = sc.ifelse(unit, self.parent_unit, self.unit)
        parent_dt = sc.ifelse(dt, 1.0)
        new.factor = time_ratio(unit1=self.unit, dt1=self.self_dt, unit2=unit, dt2=parent_dt) # Calculate the new factor
        new.update_values() # Update values
        new.v = new.values # Reset the base value
        new.factor = 1.0 # Reset everything else to be 1
        new.unit = unit
        new.self_dt = parent_dt
        new.parent_unit = unit
        new.parent_dt = parent_dt
        return new

    def to_parent(self):
        """ Create a new timepar with the same units as the parent """
        unit = self.parent_unit
        dt = self.parent_dt
        return self.to(unit=unit, dt=dt)

    def to_json(self):
        """ Export to JSON """
        attrs = ['v', 'unit', 'parent_unit', 'parent_dt', 'self_dt', 'factor']
        out = {'classname': self.__class__.__name__}
        out.update({attr:getattr(self, attr) for attr in attrs})
        out['values'] = sc.jsonify(self.values)
        return out

    # Act like a float -- TODO, add type checking
    def __add__(self, other): return self.values + other
    def __sub__(self, other): return self.values - other
    def __mul__(self, other): return self.asnew().set(v=self.v * other)
    def __pow__(self, other): return self.values ** other
    def __truediv__(self, other): return self.asnew().set(v=self.v / other)

    # ...from either side
    def __radd__(self, other): return other + self.values
    def __rsub__(self, other): return other - self.values
    def __rmul__(self, other): return self.asnew().set(v= other * self.v)
    def __rpow__(self, other): return other ** self.values
    def __rtruediv__(self, other): return other / self.values # TODO: should be a rate?

    # Handle modify-in-place methods
    def __iadd__(self, other): return self.set(v=self.v + other)
    def __isub__(self, other): return self.set(v=self.v - other)
    def __imul__(self, other): return self.set(v=self.v * other)
    def __itruediv__(self, other): return self.set(v=self.v / other)

    # Other methods
    def __neg__(self): return self.asnew().set(v=-self.v)


class dur(TimePar):
    """ Any number that acts like a duration """
    def update_values(self):
        self.values = self.v*self.factor
        return self.values


def days(v, parent_unit=None, parent_dt=None):
    """ Shortcut to ss.dur(value, units='day') """
    return dur(v=v, unit='day', parent_unit=parent_unit, parent_dt=parent_dt)


def years(v, parent_unit=None, parent_dt=None):
    """ Shortcut to ss.dur(value, units='year') """
    return dur(v=v, unit='year', parent_unit=parent_unit, parent_dt=parent_dt)


class rate(TimePar):
    """ Any number that acts like a rate; can be greater than 1 """
    def update_values(self):
        self.values = self.v/self.factor
        return self.values


def perday(v, parent_unit=None, parent_dt=None):
    """ Shortcut to ss.rate(value, units='day') """
    return rate(v=v, unit='day', parent_unit=parent_unit, parent_dt=parent_dt)


def peryear(v, parent_unit=None, parent_dt=None):
    """ Shortcut to ss.rate(value, units='year') """
    return rate(v=v, unit='year', parent_unit=parent_unit, parent_dt=parent_dt)


class time_prob(TimePar):
    """
    A probability over time (a.k.a. a cumulative hazard rate); must be >=0 and <=1.

    Note: ``ss.time_prob()`` converts one cumulative hazard rate to another with a
    different time unit. ``ss.rate_prob()`` converts an exponential rate to a cumulative
    hazard rate.
    """
    def update_values(self):
        v = self.v
        if self.isarray:
            self.values = v.copy()
            inds = np.logical_and(0.0 < v, v < 1.0)
            if inds.sum():
                rates = -np.log(1 - v[inds])
                self.values[inds] = 1 - np.exp(-rates/self.factor)
            invalid = np.logical_or(v < 0.0, 1.0 < v)
            if invalid.sum():
                errormsg = f'Invalid value {self.v} for {self}: must be 0-1. If using in a calculation, use .values instead.'
                raise ValueError(errormsg)
        else:
            if v == 0:
                self.values = 0
            elif v == 1:
                self.values = 1
            elif 0 <= v <= 1:
                rate = -np.log(1 - v)
                self.values = 1 - np.exp(-rate/self.factor)
            else:
                errormsg = f'Invalid value {self.v} for {self}: must be 0-1. If using in a calculation, use .values instead.'
                raise ValueError(errormsg)
        return self.values


class rate_prob(TimePar):
    """
    An instantaneous rate converted to a probability; must be >=0.

    Note: ``ss.time_prob()`` converts one cumulative hazard rate to another with a
    different time unit. ``ss.rate_prob()`` converts an exponential rate to a cumulative
    hazard rate.
    """
    def update_values(self):
        v = self.v
        if self.isarray:
            self.values = v.copy()
            inds = v > 0.0
            if inds.sum():
                self.values[inds] = 1 - np.exp(-v[inds]/self.factor)
            invalid = v < 0.0
            if invalid.sum():
                errormsg = f'Invalid value {self.v} for {self}: must be >=0. If using in a calculation, use .values instead.'
                raise ValueError(errormsg)
        else:
            if v == 0:
                self.values = 0
            elif v > 0:
                self.values = 1 - np.exp(-v/self.factor)
            else:
                errormsg = f'Invalid value {self.value} for {self}: must be >=0. If using in a calculation, use .values instead.'
                raise ValueError(errormsg)
        return self.values


class beta(time_prob):
    """ A container for beta (i.e. the disease transmission rate) """
    pass


#%% Unit handling

# Map different units onto the time units -- placed at the end to include the functions
unit_mapping_reverse = {
    None: [None],
    'unitless': ['unitless', 'none'],
    'day': ['d', 'day', 'days', 'perday', days, perday],
    'year': ['y', 'yr', 'year', 'years', 'peryear', years, peryear],
    'week': ['w', 'wk', 'week', 'weeks'],
    'month': ['m', 'mo', 'month', 'months'],
}
unit_mapping = {v:k for k,vlist in unit_mapping_reverse.items() for v in vlist}


def validate_unit(unit):
    """ Check that the unit is valid, and convert to a standard type """
    try:
        unit = unit_mapping[unit]
    except KeyError as E:
        errormsg = f'Invalid unit "{unit}". Valid units are:\n{sc.pp(unit_mapping_reverse, output=True)}'
        raise sc.KeyNotFoundError(errormsg) from E
    return unit


def is_unitless(unit):
    """ Check if explicitly unitless (excludes None; use not has_units() for that) """
    return unit in unit_mapping_reverse['unitless']


def has_units(unit):
    """ Check that the unit is valid and is not unitless or None """
    unit = validate_unit(unit)
    if unit is None:
        return False
    elif is_unitless(unit):
        return False
    else:
        return True



"""
Vignette 01: "Hello world" single STI
DECISIONS:
    - sim cannot be called empty, needs modules
    - module parameters can be set via the sim's par dict or via the module

TO RESOLVE:
    - what are the default network(s)?
    - what are the default demographics?
    - can the modules be added by string? (see CK.1.1)
    - how should the people parameters be changed? via the par dict (RS.1.2.3, CK.1.2.3) or via a people constructor (RA.1.2.3)?
"""
import stisim as ss

##################
# @RomeshA
##################
# RA.1.1 Simplest
sim = ss.Sim(modules=ss.HPV())
sim = ss.Sim(modules='hpv')

# RA.1.2 Par modification
# RA.1.2.1 Changing module pars via the module
sim = ss.Sim(modules=ss.HPV(eff_condoms=0.5))
sim = ss.Sim(modules=ss.HPV(pars={'eff_condoms':0.5})) # Depends what other arguments a module might need? Not a big fan of this otherwise

# RA.1.2.2 Changing module pars via the sim par dict
sim = ss.Sim(modules=ss.HPV(), pars={'hpv':{'eff_condoms':0.5}})

# RA.1.2.3 Changing people
# I am thinking here that we could have a standard people generation function and then
# if `people` is not a `People` instance, then it is instead some kwargs that are passed
# to this function. Structurally then this would be the same mechanism people could use to implement
# some of the workflows below. It also makes it very clear which parameters are module parameters and
# which are people parameters
sim = ss.Sim(modules=ss.HPV(), people={'n':10000})

# "Hello world" single STI (HPV) with a modified number of contacts
sim = ss.Sim(modules=ss.HPV(), people = {'n': 10000, 'n_partners': {'casual': 2}})

# NB. this is specific to STIsim's people generation function i.e.
def make_people(n, n_partners, ...):
    people = ss.People(n=n)
    ...
    return people

##################
# @pausz
##################
# PSL.1.1 Simplest
sim = ss.Sim(modules=ss.hiv())

##################
# @cliffckerr
##################
# CK.1.1 Simplest
sim = ss.Sim(modules='hpv')

# CK.1.2 Par modification
# CK.1.2.1 Changing module pars via the module

# Create some genotype pars
genotype_pars = {
    16: {
        'sev_fn': dict(form='logf2', k=0.25, x_infl=0, ttc=30)
    }
}
hpv = ss.HPV(genotype_pars=genotype_pars)
sim = ss.Sim(modules=hpv)

# CK.1.2.2 Changing module pars via the sim par dict
# See CK.2.2.2

# CK.1.2.3 Changing people
pars = dict(n_agents=5e3)
sim = ss.Sim(pars=pars)


##################
# @robynstuart
##################
# RS.1.1 Simplest
sim = ss.Sim(modules=ss.syphylis())

# RS.1.2 Par modification
# RS.1.2.1: via the sim's par dict
pars = dict(n_agents=100, syphilis=dict(init_prev=0.02))
sim = ss.Sim(pars=pars, modules=ss.syphilis())  # Ensure that any module keys in the parameters are added correctly

# RS.1.2.2: via the module par dict
syph_pars = dict(init_prev=0.02)
syph = ss.syphilis(pars=syph_pars)
sim = ss.Sim(modules=syph)

# RS.1.2.3 Changing people
sim = ss.Sim(people=ss.People(100), modules=ss.syphilis())
sim = ss.Sim(pars=dict(n_agents=100), modules=ss.syphilis())




"""
Import all Starsim modules
"""

# Start imports: version and settings
from .version import __version__, __versiondate__, __license__
from .settings import dtypes, options

# Optionally print the license
if options.license:
    print(__license__)

# Finish imports
from .utils            import *
from .arrays           import *
from .time             import *
from .parameters       import *
from .distributions    import *
from .people           import *
from .modules          import *
from .networks         import *
from .results          import *
from .demographics     import *
from .products         import *
from .interventions    import *
from .demographics     import *
from .disease          import *
from .diseases         import *
from .loop             import *
from .sim              import *
from .run              import *
from .calibration      import *
from .calib_components import *
from .samples          import *

# Assign the root folder
import sciris as sc
root = sc.thispath(__file__).parent

# Double-check key requirements -- should match pyproject.toml
reqs = ['sciris>=3.2.0', 'pandas>=2.0.0']
msg = f'The following dependencies for Starsim {__version__} were not met: <MISSING>.'
sc.require(reqs, message=msg)
del sc, reqs, msg # Don't keep this in the module



"""
Vignette 04: Default interventions
The individual examples are quite different so these are being kept as separate files

Features of RS vignette:
    - separation between Products and Interventions (as in HPVsim)
"""

import pandas as pd

import stisim as ss
import numpy as np
import pandas as pd
import sciris as sc

# Default interventions
#   - screening / testing
#   - antibiotics
#   - ART
#   - vaccination
#   - custom
#   - prevention (condoms, PrEP, VMMC, etc)


#############
# SCREENING/TESTING
#############

# Example 1: a dual HIV/syphilis test, defined by using a combination of two tests
test_pos = pd.read_csv('products_dx.csv')
hiv_test = ss.Product(test_positivity=test_pos[test_pos.name == 'hiv_test'])
syph_test = ss.Product(test_positivity=test_pos[test_pos.name == 'syph_test'])

# Delivery via ANC and MSM outreach
anc_screening = ss.screening(
    eligibility='pregnant',  # Pass in people attribute or indices or function that generates indices
    products=[hiv_test, syph_test],  # Pass in ss.Product() or a string to reference a pre-defined one
    prob=0.3,  # 30% coverage of all pregnant women in each year from start_year
    start_year=2015,
)

msm_screening = ss.screening(
    eligibility='msm',  # I think this is common enough that we will want to be able to target it this way
    products=[hiv_test, syph_test],
    years=np.arange(2015, 2021),  # Run a 5-year-long program
    prob=np.linspace(0, 0.3, 6),  # Linear scale-up
)

# Make a sim
pars = dict(
    n_agents=100,
    birth_rate=0.028,  # Add birth rates so that we know to capture pregnancies
    death_rate=0.008,  # Add death rate
    networks=['mf', 'msm'],  # Add an MSM network - somehow this will also add MSM as a people attribute??
)
sim = ss.Sim(pars=pars, interventions=[anc_screening, msm_screening])

# Example 2: HIV tests are offered to people after they test positive for syphilis
syph_screening = ss.screening(products=syph_test, prob=0.3, start_year=2015)
to_hiv_test = lambda sim: sim.get_intervention('syph_screening').outcomes['positive']
hiv_testing = ss.screening(eligibility=to_hiv_test, products=hiv_test, prob=0.3, start_year=2015)
sim = ss.Sim(interventions=[syph_screening, hiv_testing], modules=[ss.syphilis(), ss.hiv()])

#############
# ANTIBIOTICS
#############
products_tx = pd.read_csv('products_tx.csv')
syph_tx = ss.Product(efficacy=products_tx[products_tx.name == 'syph_tx'])
syph_treatment = ss.treatment(eligibility='syph.diagnosed', products=syph_tx, prob=0.8, start_year=2020)
sim = ss.Sim(interventions=[syph_screening, syph_tx], modules=ss.syphilis())

# Cross-cutting antibiotics
antibiotics = ss.Product(efficacy=products_tx[products_tx.name == 'antibiotics'])
gon_test = ss.Product(test_positivity=test_pos[test_pos.name == 'gon_test'])
chlam_test = ss.Product(test_positivity=test_pos[test_pos.name == 'chlam_test'])

sti_screening = ss.screening(products=[gon_test, syph_test, chlam_test], prob=0.3, start_year=2015)
sti_treatment = ss.treatment(
    eligibility=['syph.diagnosed', 'gon.diagnosed', 'chlam.diagnosed'],
    products=antibiotics, prob=0.8, start_year=2020)
sim = ss.Sim(
    interventions=[sti_screening, sti_treatment],
    modules=[ss.syphilis(), ss.gonorrhea(), ss.chlamydia()]
)

#############
# ART
#############
# Super simple example
art = ss.treatment(eligibility='hiv.diagnosed', products='art', prob=0.8, start_year=2000)

# Example where data is provided with number of people on ART per year
art_coverage = pd.read_csv('art_coverage.csv')
art = ss.treatment(eligibility='hiv.diagnosed', products='art', data=art_coverage)
# Ideally I think this should try to match the numbers


#############
# Vaccination
#############
bexsero = ss.Product()  # define a vaccine product somehow
routine_vx = ss.routine_vx(age=[9,10], sex='female', product=bexsero, prob=0.2, start_year=2025)


#############
# Custom interventions
#############
# Realistically this might be a default intervention, but let's imagine it isn't.
class condoms(ss.Intervention):
    def __init__(self, coverage=None, efficacy=None):
        self.coverage=coverage  # Coverage by network?
        self.efficacy=efficacy  # Efficacy by disease??

    def initialize(self, sim):
        if sc.checktype(self.coverage,'num'):
            self.coverage = {nkey:self.coverage for nkey in sim.networks}
        if sc.checktype(self.efficacy, 'num'):
            self.efficacy = {mkey:self.efficacy for mkey in sim.modules}

    def apply(self, sim):
        for mkey, eff in self.efficacy.items():
            for nkey, cov in self.coverage.items():
                prev_beta = sim.pars[mkey]['beta'][nkey]
                sim.pars[mkey]['beta'][nkey] = prev_beta*(1-cov) + prev_beta*cov*(1-eff)


def quarantine_syph(sim):
    """ Hypotehtical example where people with syphilis are quarantined from HIV """
    syph = sim.people.syph.infected
    sim.people.hiv.rel_sus[syph] = 0.0


class caesarian(ss.Intervention):
    """
    Give women with gonorrhea caesarians to prevent transmission from the birth canal.
    Note, this is NOT a recommended intervention!!!
    """
    def __init__(self, start_year=None, *args, **kwargs):
        super().__init__(**kwargs)
        self.start_year = start_year
        return

    def initialize(self, sim):
        super().initialize()
        self.caesareans = np.zeros(sim.npts)
        return

    def apply(self, sim):
        if sim.year >= self.start_year:
            sim.pars.gonorrhea.beta['maternal'] = [0, 0]
            self.caesareans[sim.ti] = np.count_nonzero((sim.people.ti_delivery==sim.ti) & sim.people.gonorrhea.infected)
        return



"""
Define random-number-safe distributions.
"""
import sciris as sc
import numpy as np
import numba as nb
import scipy.stats as sps
import starsim as ss
import matplotlib.pyplot as plt

__all__ = ['link_dists', 'make_dist', 'dist_list', 'Dists', 'Dist']


def str2int(string, modulo=1_000_000_000):
    """
    Convert a string to an int to use as a random seed; not for the user

    Cannot use Python's built-in hash() since it's randomized for strings. Hashlib
    (sc.sha) is 5x slower than int.from_bytes(string.encode(), byteorder='big'), but should
    only add a couple milliseconds to a typical sim.
    """
    integer = sc.sha(string, asint=True) # Hash the string to an integer
    seed = integer % modulo # Ensure a user-friendly representation user-friendly
    return seed


def link_dists(obj, sim, module=None, overwrite=False, init=False, **kwargs): # TODO: actually link the distributions to the modules! Currently this only does the opposite, but should have mod.dists as well
    """ Link distributions to the sim and the module; used in module.init() and people.init() """
    if module is None and isinstance(obj, ss.Module):
        module = obj
    dists = sc.search(obj, type=Dist, **kwargs)
    for dist in dists.values():
        dist.link_sim(sim, overwrite=overwrite)
        dist.link_module(module, overwrite=overwrite)
        if init: # Usually this is false since usually these are initialized centrally by the sim
            dist.init()
    return


def make_dist(pars=None, **kwargs):
    """ Make a distribution from a dictionary """
    pars = sc.mergedicts(pars, kwargs)
    if 'type' not in pars:
        errormsg = 'To make a distribution from a dict, one of the keys must be "type" to specify the distribution'
        raise ValueError(errormsg)
    dist_type = pars.pop('type')
    if dist_type not in dist_list:
        errormsg = f'"{dist_type}" is not a valid distribution; valid types are {dist_list}'
        raise AttributeError(errormsg)
    dist_class = getattr(ss, dist_type)
    dist = dist_class(**pars) # Actually create the distribution
    return dist


class Dists(sc.prettyobj):
    """ Class for managing a collection of Dist objects """
    def __init__(self, obj=None, *args, base_seed=None, sim=None):
        if len(args): obj = [obj] + list(args)
        self.obj = obj
        self.dists = None
        self.base_seed = base_seed
        if sim is None and obj is not None and isinstance(obj, ss.Sim):
            sim = obj
        self.sim = sim
        self.initialized = False
        return

    def init(self, obj=None, base_seed=None, sim=None, force=False):
        """
        Set the base seed, find and initialize all distributions in an object

        In practice, the object is usually a Sim, but can be anything.
        """
        if base_seed:
            self.base_seed = base_seed
        sim = sim if (sim is not None) else self.sim
        obj = obj if (obj is not None) else self.obj
        if sim is None and obj is not None and isinstance(obj, ss.Sim):
            sim = obj
        if obj is None and sim is not None:
            obj = sim
        if obj is None:
            errormsg = 'Must supply a container that contains one or more Dist objects, typically the sim'
            raise ValueError(errormsg)

        # Do not look for distributions in the people states, since they shadow the "real" states
        skip = dict(
            ids=id(sim.people._states) if sim is not None else None,
            keys='module',
        )

        # Find and initialize the distributions
        self.dists = sc.search(obj, type=Dist, skip=skip, flatten=True)
        for trace,dist in self.dists.items():
            if not dist.initialized or force:
                dist.init(trace=trace, seed=base_seed, sim=sim, force=force)

        # Confirm the seeds are unique
        self.check_seeds()
        self.initialized = True
        return self

    def check_seeds(self):
        """ Check that no two distributions share the same seed """
        checked = dict()
        for dist in self.dists.values():
            seed = dist.seed
            if seed in checked.keys():
                raise DistSeedRepeatError(checked[seed], dist)
            else:
                checked[seed] = dist
        return

    def jump(self, to=None, delta=1, force=False):
        """ Advance all RNGs, e.g. to call "to", by jumping """
        out = sc.autolist()

        # Do not jump if centralized
        if ss.options._centralized:
            return out

        for dist in self.dists.values():
            out += dist.jump(to=to, delta=delta, force=force)
        return out

    def jump_dt(self, ti=None, force=False): # Could this be simplified with jump(), or nice to have the parallel with Dist?
        """
        Advance all RNGs to the next timestep

        Args:
            ti (int): if specified, jump to this timestep (default: current sim timestep)
        """
        out = sc.autolist()

        # Do not jump if centralized
        if ss.options._centralized:
            return out

        for dist in self.dists.values():
            out += dist.jump_dt(ti=ti, force=force)
        return out

    def reset(self):
        """ Reset each RNG """
        out = sc.autolist()
        for dist in self.dists.values():
            out += dist.reset()
        return out

    def copy_to_module(self, module):
        """ Copy the Sim's Dists object to the specified module """
        matches = {key:dist for key,dist in self.dists.items() if id(dist.module) == id(module)} # Find which dists belong to this module
        if len(matches):
            new = Dists() # Create an empty Dists object
            new.__dict__.update(self.__dict__) # Shallow-copy all values over
            new.obj = module # Replace the module
            new.dists = sc.objdict(matches) # Replace the dists with a shallow copy the matching dists
            module.dists = new # Copy to the module
        else:
            new = None
        return new

class Dist:
    """
    Base class for tracking one random number generator associated with one distribution,
    i.e. one decision per timestep.

    See ss.dist_list for a full list of supported distributions.

    Although it's possible in theory to define a custom distribution (i.e., not
    one from NumPy or SciPy), in practice this is difficult. The distribution needs
    to have both a way to return random variates (easy), as well as the probability
    point function (inverse CDF). In addition, the distribution must be able to
    take a NumPy RNG as its bit generator. It's easier to just use a default Dist
    (e.g., ss.random()), and then take its output as input (i.e., quantiles) for
    whatever custom distribution you want to create.

    Args:
        dist (rv_generic): optional; a scipy.stats distribution (frozen or not) to get the ppf from
        distname (str): the name for this class of distribution (e.g. "uniform")
        name (str): the name for this particular distribution (e.g. "age_at_death")
        seed (int): the user-chosen random seed (e.g. 3)
        offset (int): the seed offset; will be automatically assigned (based on hashing the name) if None
        strict (bool): if True, require initialization and invalidate after each call to rvs()
        auto (bool): whether to auto-reset the state after each draw
        sim (Sim): usually determined on initialization; the sim to use as input to callable parameters
        module (Module): usually determined on initialization; the module to use as input to callable parameters
        kwargs (dict): parameters of the distribution

    **Examples**::

        dist = ss.Dist(sps.norm, loc=3)
        dist.rvs(10) # Return 10 normally distributed random numbers
    """
    def __init__(self, dist=None, distname=None, name=None, seed=None, offset=None,
                 strict=True, auto=True, sim=None, module=None, debug=False, **kwargs):
        # If a string is provided as "dist" but there's no distname, swap the dist and the distname
        if isinstance(dist, str) and distname is None:
            distname = dist
            dist = None
        self.dist = dist # The type of distribution
        self.distname = distname
        self.name = name
        self.pars = sc.objdict(kwargs) # The user-defined kwargs
        self.seed = seed # Usually determined once added to the container
        self.offset = offset
        self.module = module
        self.sim = sim
        self.slots = None # Created on initialization with a sim
        self.strict = strict
        self.auto = auto
        self.debug = debug

        # Auto-generated
        self.rvs_func = None # The default function to call in make_rvs() to generate the random numbers
        self.dynamic_pars = None # Whether or not the distribution has array or callable parameters
        self._pars = None # Validated and transformed (if necessary) parameters
        self._n = None # Internal variable to keep track of "n" argument (usually size)
        self._size = None # Internal variable to keep track of actual number of random variates asked for
        self._uids = None # Internal variable to track currently-in-use UIDs
        self._slots = None # Internal variable to track currently-in-use slots

        # History and random state
        self.dt_jump_size = 1000 # How much to advance the RNG for each timestep (must be larger than the number times dist is called per timestep)
        self.rng = None # The actual RNG generator for generating random numbers
        self.trace = None # The path of this object within the parent
        self.ind = 0 # The index of the RNG (usually updated on each timestep)
        self.called = 0 # The number of times the distribution has been called
        self.history = [] # Previous states
        self.ready = True
        self.initialized = False
        if not strict: # Otherwise, wait for a sim
            self.init()
        return

    def __repr__(self):
        """ Custom display to show state of object """
        j = sc.dictobj(self.to_json())
        string = f'ss.{j.classname}({j.tracestr}, {j.diststr}pars={j.pars})'
        return string

    def disp(self):
        """ Return full display of object """
        return sc.pr(self)

    def show_state(self, output=False):
        """ Show the state of the object """
        keys = ['pars', 'trace', 'offset', 'seed', 'ind', 'called', 'ready', 'state_int']
        data = {key:getattr(self,key) for key in keys}
        s = f'{self}'
        for key,val in data.items():
            s += f'\n{key:9s} = {val}'
        if output:
            return data
        else:
            print(s)
            return

    def __call__(self, n=1):
        """ Alias to self.rvs() """
        return self.rvs(n=n)

    def set(self, *args, dist=None, **kwargs):
        """ Set (change) the distribution type, or one or more parameters of the distribution """
        if dist:
            self.dist = dist
            self.process_dist()
        if args:
            if kwargs:
                errormsg = f'You can supply args or kwargs, but not both (args={len(args)}, kwargs={len(kwargs)})'
                raise ValueError(errormsg)
            else: # Convert from args to kwargs
                parkeys = list(self.pars.keys())
                for i,arg in enumerate(args):
                    kwargs[parkeys[i]] = arg
        if kwargs:
            self.pars.update(kwargs)
            self.process_pars(call=False)
        return

    @property
    def bitgen(self):
        try:    return self.rng._bit_generator
        except: return None

    @property
    def state(self):
        """ Get the current state """
        try:    return self.bitgen.state
        except: return None

    @property
    def state_int(self):
        """ Get the integer corresponding to the current state """
        try:    return self.state['state']['state']
        except: return None

    def get_state(self):
        """ Return a copy of the state """
        return self.state.copy()

    def make_history(self, reset=False):
        """ Store the current state in history """
        if reset:
            self.history = []
        self.history.append(self.get_state()) # Store the initial state
        return

    def reset(self, state=0):
        """
        Restore state, allowing the same numbers to be resampled

        Use 0 for original state, -1 for most recent state.

        **Example**::

            dist = ss.random(seed=5).init()
            r1 = dist(5)
            r2 = dist(5)
            dist.reset(-1)
            r3 = dist(5)
            dist.reset(0)
            r4 = dist(5)
            assert all(r1 != r2)
            assert all(r2 == r3)
            assert all(r4 == r1)
        """
        if not isinstance(state, dict):
            state = self.history[state]
        self.rng._bit_generator.state = state.copy()
        self.ready = True
        return self.state

    def jump(self, to=None, delta=1, force=False):
        """ Advance the RNG, e.g. to timestep "to", by jumping """

        # Do not jump if centralized # TODO: remove
        if ss.options._centralized:
            return self.state

        # Validation
        jumps = to if (to is not None) else self.ind + delta
        if self.ind >= jumps and not force:
            errormsg = f'You tried to jump the distribution "{self}" to state {jumps}, but the ' \
                       f'RNG state is already at state {self.ind}, meaning you will draw the same ' \
                        'random numbers twice. If you are sure you want to do this, set force=True.'
            raise DistSeedRepeatError(msg=errormsg)

        # Do the jumping
        self.ind = jumps
        self.reset() # First reset back to the initial state (used in case of different numbers of calls)
        if jumps: # Seems to randomize state if jumps=0
            self.bitgen.state = self.bitgen.jumped(jumps=jumps).state # Now take "jumps" number of jumps
        return self.state

    def jump_dt(self, ti=None, force=False):
        """
        Automatically jump on the next value of dt

        Args:
            ti (int): if specified, jump to this timestep (default: current module timestep plus one)
        """
        if ti is None:
            ti = self.module.t.ti + 1
        to = self.dt_jump_size*ti
        return self.jump(to=to, force=force)

    def init(self, trace=None, seed=None, module=None, sim=None, slots=None, force=False):
        """ Calculate the starting seed and create the RNG """

        if self.initialized is True and not force: # Don't warn if we have a partially initialized distribution
            msg = f'Distribution {self} is already initialized, use force=True if intentional'
            ss.warn(msg)

        # Calculate the offset (starting seed)
        self.process_seed(trace, seed)

        # Create the actual RNG
        if ss.options._centralized:
            self.rng = np.random.mtrand._rand # If _centralized, return the centralized numpy random number instance
        else:
            self.rng = np.random.default_rng(seed=self.seed)
        self.make_history(reset=True)

        # Handle the sim, module, and slots
        self.link_sim(sim)
        self.link_module(module)
        if slots is None and self.sim is not None:
            try:
                slots = self.sim.people.slot
            except Exception as E:
                warnmsg = f'Could not extract slots from sim object, is this an error?\n{E}'
                ss.warn(warnmsg)
        if slots is not None:
            self.slots = slots

        # Initialize the distribution and finalize
        self.process_dist()
        self.process_pars(call=False)
        self.ready = True
        none_dict = dict(trace=self.trace, sim=self.sim, slots=self.slots)
        none_dict = {k:v for k,v in none_dict.items() if v is None}
        if len(none_dict):
            self.initialized = 'partial'
            if self.strict:
                errormsg = f'Distribution {self} is not fully initialized, the following inputs are None:\n{none_dict.keys()}\nThis distribution may not produce valid random numbers.'
                raise RuntimeError(errormsg)
        else:
            self.initialized = True
        return self

    def link_sim(self, sim=None, overwrite=False):
        """ Shortcut for linking the sim, only overwriting an existing one if overwrite=True; not for the user """
        if (not self.sim or overwrite) and sim is not None:
            self.sim = sim
        return

    def link_module(self, module=None, overwrite=False):
        """ Shortcut for linking the module """
        if (not self.module or overwrite) and module is not None:
            self.module = module
        return

    def process_seed(self, trace=None, seed=None):
        """ Obtain the seed offset by hashing the path to this distribution; not for the user """
        unique_name = trace or self.trace or self.name
        if unique_name:
            if not self.name:
                self.name = unique_name
            self.trace = unique_name
            self.offset = str2int(unique_name) # Key step: hash the path to the distribution
        else:
            self.offset = self.offset or 0
        self.seed = self.offset + (seed or self.seed or 0)
        return

    def process_dist(self):
        """ Ensure the distribution works; not for the user """

        # Handle a SciPy distribution, if provided
        if self.dist is not None:

            # Pull out parameters of an already-frozen distribution
            if isinstance(self.dist, sps._distn_infrastructure.rv_frozen):
                if not self.initialized: # Don't do this more than once
                    self.pars = sc.dictobj(sc.mergedicts(self.pars, self.dist.kwds))

            # Convert to a frozen distribution
            if isinstance(self.dist, sps._distn_infrastructure.rv_generic):
                spars = self.process_pars(call=False)
                spars.pop('dtype', None) # Not a valid arg for SciPy distributions
                self.dist = self.dist(**spars)

            # Override the default random state with the correct one
            self.dist.random_state = self.rng

        # Set the default function for getting the rvs
        if self.distname is not None and hasattr(self.rng, self.distname): # Don't worry if it doesn't, it's probably being manually overridden
            self.rvs_func = self.distname # e.g. self.rng.uniform -- can't use the actual function because can become linked to the wrong RNG

        return

    def process_size(self, n=1):
        """ Handle an input of either size or UIDs and calculate size, UIDs, and slots; not for the user """
        if np.isscalar(n) or isinstance(n, tuple):  # If passing a non-scalar size, interpret as dimension rather than UIDs iff a tuple
            uids = None
            slots = None
            size = n
        else:
            uids = ss.uids(n)
            if len(uids):
                if self.slots is None:
                    errormsg = f'Could not find any slots in {self}. Did you remember to initialize the distribution with the sim?'
                    raise ValueError(errormsg)
                slots = self.slots[uids]
                if len(slots): # Handle case where uids is boolean
                    size = slots.max() + 1
                else:
                    size = 0
            else:
                slots = np.array([])
                size = 0

        self._n = n
        self._size = size
        self._uids = uids
        self._slots = slots
        return size, slots

    def process_pars(self, call=True):
        """ Ensure the supplied dist and parameters are valid, and initialize them; not for the user """
        self._timepar = None # Time rescalings need to be done after distributions are calculated; store the correction factor here
        self._pars = sc.cp(self.pars) # The actual keywords; shallow copy, modified below for special cases
        if call:
            self.call_pars() # Convert from function to values if needed
        spars = self.sync_pars() # Synchronize parameters between the NumPy and SciPy distributions
        return spars

    def preprocess_timepar(self, key, timepar):
        """ Method to handle how timepars are processed; not for the user. By default, scales the output of the distribution. """
        if self._timepar is None: # Store this here for later use
            self._timepar = sc.dcp(timepar) # Make a copy to avoid modifying the original
        elif timepar.factor != self._timepar.factor:
            errormsg = f'Cannot have time parameters in the same distribution with inconsistent unit/dt values: {self._pars}'
            raise ValueError(errormsg)
        self._pars[key] = timepar.v # Use the raw value, since it could be anything (including a function)
        return timepar.v # Also use this for the rest of the loop

    def convert_callable(self, key, val, size, uids):
        """ Method to handle how callable parameters are processed; not for the user """
        size_par = sc.ifelse(uids, size, ss.uids()) # Allow none size
        out = val(self.module, self.sim, size_par)
        val = np.asarray(out) # Necessary since FloatArrs don't allow slicing # TODO: check if this is correct
        self._pars[key] = val
        return val

    def call_par(self, key, val, size, uids):
        """ Check if this parameter needs to be called to be turned into an array; not for the user """
        if isinstance(val, ss.TimePar): # If it's a time parameter, transform it to a float now
            val = self.preprocess_timepar(key, val)
        if callable(val) and not isinstance(val, type): # If the parameter is callable, then call it (types can appear as callable)
            val = self.convert_callable(key, val, size, uids)
        return val

    def call_pars(self):
        """ Check if any parameters need to be called to be turned into arrays; not for the user """

        # Initialize
        size, uids = self._size, self._uids
        if self.dynamic_pars != False: # Allow "False" to prevent ever using dynamic pars (used in ss.choice())
            self.dynamic_pars = None

        # Check each parameter
        for key,val in self._pars.items():
            val = self.call_par(key, val, size, uids)

            # If it's iterable and UIDs are provided, then we need to use array-parameter logic
            if self.dynamic_pars is None and np.iterable(val) and uids is not None:
                self.dynamic_pars = True
        return

    def sync_pars(self):
        """ Perform any necessary synchronizations or transformations on distribution parameters; not for the user """
        self.update_dist_pars()
        return self._pars

    def update_dist_pars(self, pars=None):
        """ Update SciPy distribution parameters; not for the user """
        if self.dist is not None:
            pars = pars if pars is not None else self._pars
            self.dist.kwds = pars
        return

    def rand(self, size):
        """ Simple way to get simple random numbers """
        return self.rng.random(size, dtype=ss.dtypes.float) # Up to 2x faster with float32

    def make_rvs(self):
        """ Return default random numbers for scalar parameters; not for the user """
        if self.rvs_func is not None:
            rvs_func = getattr(self.rng, self.rvs_func) # Can't store this because then it references the wrong RNG after copy
            rvs = rvs_func(size=self._size, **self._pars)
        elif self.dist is not None:
            rvs = self.dist.rvs(self._size)
        else:
            errormsg = 'Dist.rvs() failed: no valid NumPy/SciPy function found in this Dist. Has it been created and initialized correctly?'
            raise ValueError(errormsg)
        return rvs

    def ppf(self, rands):
        """ Return default random numbers for array parameters; not for the user """
        rvs = self.dist.ppf(rands)
        return rvs

    def postprocess_timepar(self, rvs):
        """ Scale random variates after generation; not for the user """
        timepar = self._timepar # Shorten
        self._timepar = None # Remove the timepar which is no longer needed
        timepar.v = rvs # Replace the base value with the random variates
        timepar.update_cached() # Recalculate the factor and values with the time scaling
        rvs = timepar.values # Replace the rvs with the scaled version
        if isinstance(rvs, np.ndarray): # This can be false when converting values for a Bernoulli distribution (in which case rvs are actually dist parameters)
            rvs = rvs.astype(rvs.dtype) # Replace the random variates with the scaled version, and preserve type
        return rvs

    def rvs(self, n=1, reset=False):
        """
        Get random variates -- use this!

        Args:
            n (int/tuple/arr): if an int or tuple, return this many random variates; if an array, treat as UIDs
            reset (bool): whether to automatically reset the random number distribution state after being called
        """
        # Check for readiness
        if not self.initialized:
            raise DistNotInitializedError(self)
        if not self.ready and self.strict and not ss.options._centralized:
            raise DistNotReadyError(self)

        # Figure out size, UIDs, and slots
        size, slots = self.process_size(n)

        # Check if size is 0, then we can return
        if size == 0:
            return np.array([], dtype=ss.dtypes.int) # int dtype allows use as index, e.g. when filtering
        elif isinstance(size, ss.uids) and self.initialized == 'partial': # This point can be reached if and only if strict=False and UIDs are used as input
            errormsg = f'Distribution {self} is only partially initialized; cannot generate random numbers to match UIDs'
            raise ValueError(errormsg)

        # Store the state
        self.make_history() # Store the pre-call state

        # Check if any keywords are callable -- parameters shouldn't need to be reprocessed otherwise
        self.process_pars()

        # Actually get the random numbers
        if self.dynamic_pars:
            rands = self.rand(size)[slots] # Get random values
            rvs = self.ppf(rands) # Convert to actual values via the PPF
        else:
            rvs = self.make_rvs() # Or, just get regular values
            if self._slots is not None:
                rvs = rvs[self._slots]

        # Scale by time if needed
        if self._timepar is not None:
            rvs = self.postprocess_timepar(rvs)

        # Tidy up
        self.called += 1
        if reset:
            self.reset(-1)
        elif self.auto: # TODO: check
            self.jump()
        elif self.strict:
            self.ready = False
        if self.debug:
            tistr   = f'on ti={self.module.ti} ' if self.module else ''
            sizestr = f'with size={self._size}, '
            slotstr = f'Σ(slots)={self._slots.sum()}, ' if self._slots else '<no slots>, '
            rvstr   = f'Σ(rvs)={rvs.sum():0.2f}, |rvs|={rvs.mean():0.4f}'
            pre_state = str(self.history[-1]['state']['state'])[-5:]
            post_state = str(self.state_int)[-5:]
            statestr = f"state {pre_state}→{post_state}"
            print(f'Debug: {self} called {tistr}{sizestr}{slotstr}{rvstr}, {statestr}')
            assert pre_state != post_state # Always an error if the state doesn't change after drawing random numbers

        return rvs

    def to_json(self):
        """ Return a dictionary representation of the Dist """
        tracestr = '<no trace>' if self.trace is None else f"{self.trace}"
        classname = self.__class__.__name__
        diststr = ''
        if classname == 'Dist':
            if self.dist is not None:
                try:
                    diststr = f'dist={self.dist.name}, '
                except:
                    try: # What is wrong with you, SciPy -- after initialization, it moves here
                        diststr = f'dist={self.dist.dist.name}, '
                    except:
                        diststr = f'dist={type(self.dist)}'
            elif self.distname is not None:
                diststr = f'dist={self.distname}, '
        out = dict(
            type = 'Dist',
            classname = classname,
            tracestr = tracestr,
            diststr = diststr,
            pars = dict(self.pars),
        )
        return out


    def plot_hist(self, n=1000, bins=None, fig_kw=None, hist_kw=None):
        """ Plot the current state of the RNG as a histogram """
        plt.figure(**sc.mergedicts(fig_kw))
        rvs = self.rvs(n)
        self.reset(-1) # As if nothing ever happened
        plt.hist(rvs, bins=bins, **sc.mergedicts(hist_kw))
        plt.title(str(self))
        plt.xlabel('Value')
        plt.ylabel(f'Count ({n} total)')
        return rvs


#%% Specific distributions

# Add common distributions so they can be imported directly; assigned to a variable since used in help messages
dist_list = ['random', 'uniform', 'normal', 'lognorm_ex', 'lognorm_im', 'expon', 'poisson', 'nbinom',
             'weibull', 'gamma', 'constant', 'randint', 'rand_raw', 'bernoulli', 'choice', 'histogram']
__all__ += dist_list
__all__ += ['multi_random'] # Not a dist in the same sense as the others (e.g. same tests would fail)


class random(Dist):
    """ Random distribution, with values on the interval (0, 1) """
    def __init__(self, **kwargs):
        super().__init__(distname='random', dtype=ss.dtypes.float, **kwargs)
        return

    def ppf(self, rands):
        return rands


class uniform(Dist):
    """
    Uniform distribution, values on interval (low, high)

    Args:
        low (float): the lower bound of the distribution (default 0.0)
        high (float): the upper bound of the distribution (default 1.0)
    """
    def __init__(self, low=None, high=None, **kwargs):
        if high is None and low is not None: # One argument, swap
            high = low
            low = 0.0
        if low is None:
            low = 0.0
        if high is None:
            high = 1.0
        super().__init__(distname='uniform', low=low, high=high, **kwargs)
        return

    def make_rvs(self):
        """ Specified here because uniform() doesn't take a dtype argument """
        p = self._pars
        rvs = self.rand(self._size) * (p.high - p.low) + p.low
        return rvs

    def ppf(self, rands):
        p = self._pars
        rvs = rands * (p.high - p.low) + p.low
        return rvs


class normal(Dist):
    """
    Normal distribution

    Args:
        loc (float): the mean of the distribution (default 0.0)
        scale (float) the standard deviation of the distribution (default 1.0)

    """
    def __init__(self, loc=0.0, scale=1.0, **kwargs): # Does not accept dtype
        super().__init__(distname='normal', dist=sps.norm, loc=loc, scale=scale, **kwargs)
        return


class lognorm_im(Dist):
    """
    Lognormal distribution, parameterized in terms of the "implicit" (normal)
    distribution, with mean=loc and std=scale (see lognorm_ex for comparison).

    Note: the "loc" parameter here does *not* correspond to the mean of the resulting
    random variates!

    Args:
        mean (float): the mean of the underlying normal distribution (not this distribution) (default 0.0)
        sigma (float): the standard deviation of the underlying normal distribution (not this distribution) (default 1.0)

    **Example**::

        ss.lognorm_im(mean=2, sigma=1, strict=False).rvs(1000).mean() # Should be roughly 10
    """
    def __init__(self, mean=0.0, sigma=1.0, **kwargs): # Does not accept dtype
        super().__init__(distname='lognormal', dist=sps.lognorm, mean=mean, sigma=sigma, **kwargs)
        return

    def sync_pars(self, call=True):
        """ Translate between NumPy and SciPy parameters """
        if call:
            self.call_pars()
        p = self._pars
        spars = sc.dictobj()
        spars.s = p.sigma
        spars.scale = np.exp(p.mean)
        spars.loc = 0
        self.update_dist_pars(spars)
        return spars

    def preprocess_timepar(self, key, val):
        """ Not valid since incorrect time units """
        errormsg = f'Cannot use timepars with a lognorm_im distribution ({self}) since its units are not time. Use lognorm_ex instead.'
        raise NotImplementedError(errormsg)


class lognorm_ex(Dist):
    """
    Lognormal distribution, parameterized in terms of the "explicit" (lognormal)
    distribution, with mean=mean and std=std for this distribution (see lognorm_im for comparison).
    Note that a mean ≤ 0.0 is impossible, since this is the parameter of the distribution
    after the log transform.

    Args:
        mean (float): the mean of this distribution (not the underlying distribution) (default 1.0)
        std (float): the standard deviation of this distribution (not the underlying distribution) (default 1.0)

    **Example**::

        ss.lognorm_ex(mean=2, std=1, strict=False).rvs(1000).mean() # Should be close to 2
    """
    def __init__(self, mean=1.0, std=1.0, **kwargs): # Does not accept dtype
        super().__init__(distname='lognormal', dist=sps.lognorm, mean=mean, std=std, **kwargs)
        return

    def convert_ex_to_im(self):
        """
        Lognormal distributions can be specified in terms of the mean and standard
        deviation of the "explicit" lognormal distribution, or the "implicit" normal distribution.
        This function converts the parameters from the lognormal distribution to the
        parameters of the underlying (implicit) distribution, which are the form expected by NumPy's
        and SciPy's lognorm() distributions.
        """
        self.call_pars() # Since can't work with functions
        p = self._pars
        mean = p.pop('mean')
        std = p.pop('std')
        if np.isscalar(mean) and mean <= 0:
            errormsg = f'Cannot create a lognorm_ex distribution with mean≤0 (mean={mean}); did you mean to use lognorm_im instead?'
            raise ValueError(errormsg)
        std2 = std**2
        mean2 = mean**2
        sigma_im = np.sqrt(np.log(std2/mean2 + 1)) # Computes std for the underlying normal distribution
        mean_im  = np.log(mean2 / np.sqrt(std2 + mean2)) # Computes the mean of the underlying normal distribution
        p.mean = mean_im
        p.sigma = sigma_im
        return mean_im, sigma_im

    def sync_pars(self):
        """ Convert from overlying to underlying parameters, then translate to SciPy """
        self.convert_ex_to_im()
        spars = lognorm_im.sync_pars(self, call=False) # Borrow sync_pars from lognorm_im
        return spars


class expon(Dist):
    """
    Exponential distribution

    Args:
        scale (float): the scale of the distribution (default 1.0)

    """
    def __init__(self, scale=1.0, **kwargs):
        super().__init__(distname='exponential', dist=sps.expon, scale=scale, **kwargs)
        return


class poisson(Dist): # TODO: does not currently scale correctly with dt
    """
    Poisson distribution

    Args:
        lam (float): the scale of the distribution (default 1.0)
    """
    def __init__(self, lam=1.0, **kwargs):
        super().__init__(distname='poisson', dist=sps.poisson, lam=lam, **kwargs)
        return

    def sync_pars(self):
        """ Translate between NumPy and SciPy parameters """
        spars = dict(mu=self._pars.lam)
        self.update_dist_pars(spars)
        return spars

    def preprocess_timepar(self, key, timepar):
        """ Try to update the timepar before calculating array parameters, but raise an exception if this isn't possible """
        try:
            timepar.update_cached()
        except Exception as E:
            errormsg = f'Could not process timepar {timepar} for {self}. Note that Poisson distributions are not compatible with both callable parameters and timepars, since this would change the shape in an unknowable way.'
            raise ValueError(errormsg) from E

        self._pars[key] = timepar.values # Use the raw value, since it could be anything (including a function)
        return timepar.values # Also use this for the rest of the loop


class nbinom(Dist):
    """
    Negative binomial distribution

    Args:
        n (float): the number of successes, > 1 (default 1.0)
        p (float): the probability of success in [0,1], (default 0.5)

    """
    def __init__(self, n=1, p=0.5, **kwargs):
        super().__init__(distname='negative_binomial', dist=sps.nbinom, n=n, p=p, **kwargs)
        return


class randint(Dist):
    """
    Random integer distribution, on the interval [low, high)

    Args:
        low (int): the lower bound of the distribution (default 0)
        high (int): the upper bound of the distribution (default of maximum integer size: 9,223,372,036,854,775,807)
        allow_time (bool): allow time parameters to be specified as high/low values (disabled by default since introduces rounding error)
    """
    def __init__(self, *args, low=None, high=None, dtype=ss.dtypes.rand_int, allow_time=False, **kwargs):
        # Handle input arguments # TODO: reconcile with how this is handled in uniform()
        self.allow_time = allow_time
        if len(args):
            if len(args) == 1:
                high = args[0]
            elif len(args) == 2:
                low,high = args
            else:
                errormsg = f'ss.randint() takes one or two arguments, not {len(args)}'
                raise ValueError(errormsg)
        if low is None:
            low = 0
        if high is None:
            high = np.iinfo(ss.dtypes.rand_int).max

        if ss.options._centralized: # randint because we're accessing via numpy.random
            super().__init__(distname='randint', low=low, high=high, dtype=dtype, **kwargs)
        else: # integers instead of randint because interfacing a numpy.random.Generator
            super().__init__(distname='integers', low=low, high=high, dtype=dtype, **kwargs)
        return

    def ppf(self, rands):
        p = self._pars
        rvs = rands * (p.high + 1 - p.low) + p.low
        rvs = rvs.astype(self.dtype)
        return rvs

    def preprocess_timepar(self, key, timepar):
        """ Not valid due to a rounding error """
        if self.allow_time:
            return super().preprocess_timepar(key, timepar)
        else:
            errormsg = f'Cannot use timepars with a randint distribution ({self}) since the values may be rounded incorrectly. Use uniform() instead and convert to int yourself, or set allow_time=True if you really want to do this.'
            raise NotImplementedError(errormsg)


class rand_raw(Dist):
    """
    Directly sample raw integers (uint64) from the random number generator.
    Typicaly only used with ss.combine_rands().
    """
    def make_rvs(self):
        if ss.options._centralized:
            return self.rng.randint(low=0, high=np.iinfo(np.uint64).max, dtype=np.uint64, size=self._size)
        else:
            return self.bitgen.random_raw(self._size) # TODO: figure out how to make accept dtype, or check speed


class weibull(Dist):
    """
    Weibull distribution (specifically, scipy.stats.weibull_min)

    Args:
        c (float): the shape parameter, sometimes called k (default 1.0)
        loc (float): the location parameter, which shifts the position of the distribution (default 0.0)
        scale (float): the scale parameter, sometimes called λ (default 1.0)
    """
    def __init__(self, c=1.0, loc=0.0, scale=1.0, **kwargs):
        super().__init__(distname='weibull', dist=sps.weibull_min, c=c, loc=loc, scale=scale, **kwargs)
        return

    def make_rvs(self):
        """ Use SciPy rather than NumPy to include the scale parameter """
        rvs = self.dist.rvs(self._size)
        return rvs


class gamma(Dist):
    """
    Gamma distribution (specifically, scipy.stats.gamma)

    Args:
        a (float): the shape parameter, sometimes called k (default 1.0)
        loc (float): the location parameter, which shifts the position of the distribution (default 0.0)
        scale (float): the scale parameter, sometimes called θ (default 1.0)
    """
    def __init__(self, a=1.0, loc=0.0, scale=1.0, **kwargs):
        super().__init__(distname='gamma', dist=sps.gamma, a=a, loc=loc, scale=scale, **kwargs)
        return

    def make_rvs(self):
        """ Use SciPy rather than NumPy to include the scale parameter """
        rvs = self.dist.rvs(self._size)
        return rvs


class constant(Dist):
    """
    Constant (delta) distribution: equivalent to np.full()

    Args:
        v (float): the value to return
    """
    def __init__(self, v=0.0, **kwargs):
        super().__init__(distname='const', v=v, **kwargs)
        return

    def make_rvs(self):
        return np.full(self._size, self._pars.v)

    def ppf(self, rands): # NB: don't actually need to use random numbers here, but not worth the complexity of avoiding this
        return np.full(rands.shape, self._pars.v)


class bernoulli(Dist):
    """
    Bernoulli distribution: return True or False with the specified probability (which can be an array)

    Unlike other distributions, Bernoulli distributions have a filter() method,
    which returns elements of the array that return True.

    Args:
        p (float): the probability of returning True (default 0.5)
    """
    def __init__(self, p=0.5, **kwargs):
        super().__init__(distname='bernoulli', p=p, **kwargs)
        return

    def make_rvs(self):
        rvs = self.rand(self._size) < self._pars.p # 3x faster than using rng.binomial(1, p, size)
        return rvs

    def ppf(self, rands):
        rvs = rands < self._pars.p
        return rvs

    def filter(self, uids=None, both=False):
        """ Return UIDs that correspond to True, or optionally return both True and False """
        if uids is None:
            uids = self.sim.people.auids # All active UIDs
        elif isinstance(uids, (ss.BoolArr, ss.IndexArr)):
            uids = uids.uids

        bools = self.rvs(uids)
        if both:
            return uids[bools], uids[~bools]
        else:
            return uids[bools]

    def split(self, uids=None):
        """ Alias to filter(uids, both=True) """
        return self.filter(uids=uids, both=True)

    def call_par(self, key, val, size, uids):
        """ Reverse the usual order of processing so callable is processed first, and then the timepar conversion """
        is_timepar = isinstance(val, ss.TimePar)

        if is_timepar: # If it's a time parameter, pull out the value
            timepar = sc.dcp(val) # Rename to make more sense within the context of this method
            val = timepar.v # Pull out the base value; we'll deal with the transformation later
            self._timepar = timepar # This is used, then destroyed, by postprocess_timepar() below
            if isinstance(timepar, ss.dur): # Validation
                errormsg = f'Bernoulli distributions can only be used with ss.time_prob() or ss.rate(), not {timepar}'
                raise TypeError(errormsg)

        # As normal: if the parameter is callable, then call it (types can appear as callable)
        if callable(val) and not isinstance(val, type):
            val = self.convert_callable(key, val, size, uids)

        # Process as a timepar
        if is_timepar:
            val = self.postprocess_timepar(val) # Note: this is processing the parameter rather than the rvs as usual

        # Store in the parameters and return
        self._pars[key] = val
        return val


class choice(Dist):
    """
    Random choice between discrete options (note: dynamic parameters not supported)

    Args:
        a (int or array): the number of choices, or the choices themselves (default 2)
        p (array): if supplied, the probability of each choice (default, 1/a for a choices)

    **Examples**::

        # Simulate 10 die rolls
        ss.choice(6, strict=False)(10) + 1

        # Choose between specified options each with a specified probability (must sum to 1)
        ss.choice(a=[30, 70], p=[0.3, 0.7], strict=False)(10)

    Note: although Bernoulli trials can be generated using a=2, it is much faster
    to use ss.bernoulli() instead.
    """
    def __init__(self, a=2, p=None, **kwargs):
        super().__init__(distname='choice', a=a, p=p, **kwargs)
        self.dynamic_pars = False # Set to false since array arguments don't imply dynamic pars here
        return

    def ppf(self, rands):
        """ Shouldn't actually be needed since dynamic pars not supported """
        pars = self._pars
        if np.isscalar(pars.a):
            pars.a = np.arange(pars.a)
        pcum = np.cumsum(pars.p)
        inds = np.searchsorted(pcum, rands)
        rvs = pars.a[inds]
        return rvs

    def preprocess_timepar(self, key, timepar):
        """ Not valid since does not scale with time """
        errormsg = f'Cannot use timepars with a choice distribution ({self}) since its units are not time. Convert output to time units instead.'
        raise NotImplementedError(errormsg)


class histogram(Dist):
    """
    Sample from a histogram with defined bins

    Note: unlike other distributions, the parameters of this distribution can't
    be modified after creation.

    Args:
        values (array): the probability (or count) of each bin
        bins (array): the edges of each bin
        density (bool): treat the histogram as a density instead of counts; only matters with unequal bin widths, see numpy.histogram and scipy.stats.rv_histogram for more information
        data (array): if supplied, compute the values and bin edges using this data and np.histogram() instead

    Note: if the length of bins is equal to the length of values, they will be
    interpreted as left bin edges, and one additional right-bin edge will be added
    based on the difference between the last two bins (e.g. if the last two bins are
    40 and 50, the final right edge will be added at 60). If no bins are supplied,
    then they will be created as integers matching the length of the values.

    The values can be supplied in either normalized (sum to 1) or un-normalized
    format.

    **Examples**::

        # Sample from an age distribution
        age_bins = [0,    10,  20,  40,  65, 100]
        age_vals = [0.1, 0.1, 0.3, 0.3, 0.2]
        h1 = ss.histogram(values=age_vals, bins=age_bins, strict=False)
        h1.plot_hist()

        # Create a histogram from data
        data = np.random.randn(10_000)*2+5
        h2 = ss.histogram(data=data, strict=False)
        h2.plot_hist(bins=100)
    """
    def __init__(self, values=None, bins=None, density=False, data=None, **kwargs):
        if data is not None:
            if values is not None:
                errormsg = 'You can supply values or data, but not both'
                raise ValueError(errormsg)
            kw = {'bins':bins} if bins is not None else {} # Hack to not overwrite default bins value
            values, bins = np.histogram(data, **kw)
        else:
            if values is None:
                values = [1.0] # Uniform distribution
            values = sc.toarray(values)
            if bins is None:
                bins = np.arange(len(values)+1)
            bins = sc.toarray(bins)
        if len(bins) == len(values): # Append a final bin, if necessary
            delta = bins[-1] - bins[-2]
            bins = np.append(bins, bins[-1]+delta)
        vsum = values.sum()
        if vsum != 1.0:
            values = values / vsum
        dist = sps.rv_histogram((values, bins), density=density) # Create the SciPy distribution
        super().__init__(dist=dist, distname='histogram', **kwargs)
        self.dynamic_pars = False # Set to false since array arguments don't imply dynamic pars here
        return


class multi_random(sc.prettyobj):
    """
    A class for holding two or more ss.random() distributions, and generating
    random numbers linked to each of them. Useful for e.g. pairwise transmission
    probabilities.

    See ss.combine_rands() for the manual version; in almost all cases this class
    should be used instead.

    Usage:
        multi = ss.multi_random('source', 'target')
        rvs = multi.rvs(source_uids, target_uids)
    """
    def __init__(self, names, *args, **kwargs):
        names = sc.mergelists(names, args)
        self.dists = [ss.random(name=name, **kwargs) for name in names]
        return

    def __len__(self):
        return len(self.dists)

    def init(self, *args, **kwargs):
        """ Not usually needed since each dist will handle this automatically; for completeness only """
        for dist in self.dists: dist.init(*args, **kwargs)
        return

    def reset(self, *args, **kwargs):
        """ Not usually needed since each dist will handle this automatically; for completeness only """
        for dist in self.dists: dist.reset(*args, **kwargs)
        return

    def jump(self, *args, **kwargs):
        """ Not usually needed since each dist will handle this automatically; for completeness only """
        for dist in self.dists: dist.jump(*args, **kwargs)
        return

    @staticmethod
    @nb.njit(fastmath=True, parallel=False, cache=True) # Numba is 3x faster, but disabling parallel for efficiency
    def combine_rvs(rvs_list, int_type, int_max):
        """ Combine inputs into one number """
        # Combine using bitwise-or
        rand_ints = rvs_list[0].view(int_type)
        for rand_floats in rvs_list[1:]:
            rand_ints2 = rand_floats.view(int_type)
            rand_ints = np.bitwise_xor(rand_ints*rand_ints2, rand_ints-rand_ints2)

        # Normalize
        rvs = rand_ints / int_max
        return rvs

    def rvs(self, *args):
        """ Get random variates from each of the underlying distributions and combine them efficiently """
        # Validation
        n_args = len(args)
        n_dists = len(self)
        if n_args != len(self):
            errormsg = f'Number of UID lists supplied ({n_args}) does not match number of distributions ({n_dists})'
            raise ValueError(errormsg)

        rvs_list = [dist.rvs(arg) for dist,arg in zip(self.dists, args)]
        int_type = ss.dtypes.rand_uint
        int_max = np.iinfo(int_type).max
        rvs = self.combine_rvs(rvs_list, int_type, int_max)
        return rvs




#%% Dist exceptions

class DistNotInitializedError(RuntimeError):
    """ Raised when Dist object is called when not initialized. """
    def __init__(self, dist=None, msg=None):
        if msg is None:
            msg = f'{dist} has not been initialized; please set strict=False when creating the distribution, or call dist.init()'
        super().__init__(msg)
        return


class DistNotReadyError(RuntimeError):
    """ Raised when a Dist object is called without being ready. """
    def __init__(self, dist=None, msg=None):
        if msg is None:
            msg = f'{dist} is not ready. This is likely caused by calling a distribution multiple times in a single step. Call dist.jump() to reset.'
        super().__init__(msg)
        return


class DistSeedRepeatError(RuntimeError):
    """ Raised when a Dist object shares a seed with another """
    def __init__(self, dist1=None, dist2=None, msg=None):
        if msg is None:
            msg = f'A common seed was found between {dist1} and {dist2}. This is likely caused by incorrect initialization of the parent Dists object.'
        super().__init__(msg)
        return
